unit baseoverlay;

interface
{$i vwrdefines.inc}

Uses candtype,dglOpenGL,Classes,Graphics,windows,sysutils;

const
  nrOverlayLayers = 5;
  GLOverlaymagic  = $72095725;
  GLOverlayVersion= $10002;
  MaxRegions      = 30;
  MaxMarkedObjects= 30;
  Maxblobrect     = 30;

type
  TPointGLFloat= record
                  x,y:GLFloat;
                  end;


  TGLOverlayHeader = record
                       magic   : integer;
                       hdrsize : integer;
                       version : integer;
                       layers  : integer;
                       theend  : record end;   // 0 sized field to mark end of guaranteed hdr
                       // new fields.
                       end;

  TRegionArray = Array[0..maxregions-1] of integer;

  ///////////////////
  /// GL VERSION
  ///////////////////

  { These are the basic primitives and higher constructs
    for holding the draw data of various types. Currently we have:
    1. GL_LINES (Xpx line, defined by 2 points (not lineto style), width defined globally
    2. GL_QUADS (filled rect, defined by 4 points
    3. GL_TRIANGLES (filled triangle, defined by 3 points, used for filled ellipses

    For each of these, there's also colors info, per vertex/point
  }

  TGLVertex = array[1..2] of GLfloat; // basic 2d vertex, no Z for now
  TGLLine = array[1..2] of TGLVertex; // line from 2 vertices
  TGLTriangle = array[1..3] of TGLVertex; // triangle from 3 vertices
  TGLRect = array[1..4] of TGLVertex; // rect vertex definitions

  TGLColor = array[1..4] of Byte; // 3 singles for RGB
  TGLLineColor = array[1..2] of TGLColor; // 2 colors, 1 per line vertex
  TGLTriangleColor = array[1..3] of TGLColor; // 3 colors, 1 per triangle vertex
  TGLRectColor = array[1..4] of TGLColor; // 4 colors, 1 per rect vertex

  { This is the internal status of the overlay.
    It was more useful in the old TGLOverlayDL with display lists
    than with the VBO one, but I decided to leave it in anyhow.
    None means "not yet used in any way"
    Filling means that some primitives were drawn, but not flushed
    Flushed means "ready for drawing"
  }
  TGLOverlayStatus = (glsNone, glsFilling, glsFlushed);
  TOverlayObjectType = (ovlRect,ovlRectfilled,ovlTriangle,ovlLine,ovlThickline,ovlStippleline,ovlText,ovlCirclefilled,ovlCircleEmpty);
  TOverlayObjectArray = Array[TOverlayObjectType] of integer;
  TMarkedObject = record
                    objindex: integer;
                    objtype : TOverlayObjectType;
                    number  : word;
                  end;

  { Base class for the overlay. Holds common stuff
    Owner needs to be known because of wglMakeCurrent call
    which is required to be done before ANY GL operations
    AND is completely thread unsafe. It basically tells
    GL "I'm working with this context now". }
  TOverlaydata=class;
  TOverlayLayerData = Class
                    width,height:integer;
                    FRects: array of TGLRect;
                    FRectColors: array of TGLRectColor;
                    FRectCount: GLint;

                    FTriangles: array of TGLTriangle;
                    FTriangleColors: array of TGLTriangleColor;
                    FTriangleCount: GLint;

                    FLines: array of TGLLine;
                    FLinecolors: array of TGLLinecolor;
                    FLineCount: GLint;

                    FThickLines: array of TGLLine;
                    FThickLinecolors: array of TGLLinecolor;
                    FThickLineCount: GLint;

                    FStippleLines: array of TGLLine;
                    FStippleLinecolors: array of TGLLinecolor;
                    FStippleLineCount: GLint;

                    FTextColors: array of TGLColor;
                    FTexts: TStringList;
                    FTextVertices: array of TGLVertex;

                    FAreaBounderyCount : integer;
                    FAreaBounderies : array[0..maxRegions-1] of TOverlayObjectArray;
                    FMarkedObjectCount : integer;
                    FMarkedObject   : array[0..MaxMarkedObjects-1] of TMarkedObject;

                    procedure Clear;
                    constructor Create;
                    destructor Destroy; override;

                    procedure InflateRects;
                    procedure InflateLines;
                    procedure InflateTriangles;
                    procedure InflateThickLines;
                    procedure InflateStippleLines;
                    procedure assign(x:TOverlaylayerdata);
                    procedure InflateTextArrays;
                    { Concrete implementation of DrawText, directly in the base class.
                      x, y define position, color has alpha value in first 16 bits.
                      "thread-almost-safe" }
                    procedure DrawText(const x, y: GLfloat; const aText: string; const aColor: TColor);
                    procedure DrawTriangle(x1, y1, x2, y2, x3, y3: GLfloat; const aColor: TColor;
                                               const IsFilled: Boolean = False);                  {$ifndef ver150} inline; {$endif}
                    procedure IntDrawTriangle(x1, y1, x2, y2, x3, y3: GLfloat; const aColor: TColor);{$ifndef ver150} inline; {$endif}

                    { Base method for DrawLine defined by 2 points in turn defined by x1, y1 and
                      x2, y2 GLfloats/singles. "thread-almost-safe" in VBO version }
                    procedure DrawLine(x1, y1, x2, y2: GLfloat; const aColor: TColor);
                    procedure intDrawLine(x1, y1, x2, y2: GLfloat; const aColor: TColor);
                    { Base method for DrawThickLine defined by 2 points in turn defined by x1, y1 and
                      x2, y2 GLfloats/singles. "thread-almost-safe" in VBO version }
                    procedure DrawThickLine(x1, y1, x2, y2: GLfloat; const aColor: TColor);
                    { Base method for DrawRect defined by 2 points in turn defined by x1, y1 and
                      x2, y2 GLfloats/singles. "thread-almost-safe" in VBO version }
                    procedure DrawStippleLine(x1, y1, x2, y2: GLfloat; const aColor: TColor);
                    { Base method for DrawRect defined by 2 points in turn defined by x1, y1 and
                      x2, y2 GLfloats/singles. "thread-almost-safe" in VBO version }
                    procedure DrawRect(x1, y1, x2, y2: GLfloat; const aColor: TColor;
                                          const IsFilled: Boolean = False);
                    { Base method for DrawEllipse defined by 1 center point and width/height (half)
                      Ellipses are very costy, each has 360*2 structures of base type.
                      "thread-almost-safe" in VBO version }
                    procedure DrawEllipse(x, y, w, h: GLfloat; const aColor: TColor;
                                          const IsFilled: Boolean = False);
                    { Base method for polygon. Draw n points. Filling NOT yet supported.   }
                    procedure DrawPolygon(const pts:array of TPointGLFloat; const aColor: TColor;const IsFilled: Boolean = False);
                    procedure loadfromstream(t:tstream;vers:integer);
                    procedure savetostream(t:tstream);
                    procedure markobject(t:TOverlayObjectType;nr:integer=1);
                    procedure colorselection(col:tcolor);
                 end;


  TOverlayData = class
                    private
                      fwidth,fheight: integer;
                      // blobs are things like spots that we must be able to iterate and focus on.
                      fblobcount : integer;
                      fblobrects : array[0..Maxblobrect-1] of TRect;   // the rough bounding box of the blob.
                      fbloblevel : array[0..Maxblobrect-1] of Integer; // room for a "level/category " for each blob.
                    public
                      Layers : array[1..nroverlaylayers] of TOverlayLayerData;
                    procedure markblob(x1,y1,x2,y2:integer;level:integer);
                    procedure assign(x:TOverlaydata);
                    constructor Create;
                    procedure Clear;
                    destructor Destroy; override;
                    procedure setwidth(w:integer);
                    procedure setheight(h:integer);
                    procedure loadfromstream(t:tstream);
                    procedure savetostream(t:tstream);
                    function getblob(ind:integer):trect;
                    property blob[i:integer] :TRect read getblob;
                    property blobcount : integer read fblobcount;
                    property width : integer read fwidth write setwidth;
                    property height:integer read fheight write setheight;
                 end;
  TBasicOverlay = TOverlayData;

implementation

uses basehistory, Types, ctypes;

{ For faster calculation }
const
  DEG2RAD: GLfloat = System.PI / 180;

procedure TOverlayLayerData.DrawText(const x, y: GLfloat; const aText: string;
  const aColor: TColor);
var
  Index: Integer;
begin

  FTexts.Add(aText);

  if FTexts.Count >= Length(FTextColors) then
    InflateTextArrays;
  Index := FTexts.Count - 1;

  FTextVertices[Index][1] := x/width;
  FTextVertices[Index][2] := y/height;
  cardinal(pcolor(@FTextColors[index][1])^):=(acolor and $00FFFFFF) or $FF000000;
end;

procedure TOverlayLayerData.DrawPolygon(const pts:array of TPointGLFloat; const aColor: TColor;const IsFilled: Boolean = False);
var i,l : integer;
begin
  l:=length(pts);
  if l<2 then exit;
  if l=2 then
    begin
      drawline(pts[0].x,pts[0].y,pts[1].x,pts[1].y,acolor);
      exit;
    end;
  for i:=0 to l-2 do
    begin
       drawline(pts[i].x,pts[i].y,pts[i+1].x,pts[i+1].y,acolor);
    end;
  drawline(pts[l-1].x,pts[l-1].y,pts[0].x,pts[0].y,acolor);
end;

procedure TOverlayLayerData.InflateTextArrays;
const
  DEF_SIZE = 100;
begin
  SetLength(FTextColors, Length(FTextColors) + DEF_SIZE);
  SetLength(FTextVertices, Length(FTextVertices) + DEF_SIZE);
end;

procedure TOverlayLayerData.Clear;
begin
 {  inherited;}

   FTexts.Clear;
 {end inherited}

  FRectCount := 0;
  FLineCount := 0;
  FThickLineCount := 0;
  FStippleLineCount := 0;
  FTriangleCount := 0;
  FMarkedObjectCount:=0;


//  FListStatus := glsFilling;
end;

procedure TOverlayLayerData.assign(x:TOverlaylayerdata);

begin
  frectcount:=x.frectcount;
  frects:=copy(x.frects);
  FRectColors:=copy(x.frectcolors);

  ftriangles:=copy(x.FTriangles);
  ftrianglecolors:=copy(x.FTrianglecolors);
  ftrianglecount:=x.FTrianglecount;

  flines:=copy(x.FLines);
  flinecolors:=copy(x.FLinecolors);
  flinecount:=x.FLineCount;

  fthicklines:=copy(x.Fthicklines);
  fthicklinecolors:=copy(x.Fthicklinecolors);
  fthicklinecount:=x.FthicklineCount;

  fstipplelines:=copy(x.Fstipplelines);
  fstipplelinecolors:=copy(x.Fstipplelinecolors);
  fstipplelinecount:=x.FstipplelineCount;

  FTexts.assign(x.ftexts);
  FTextColors:=copy(x.ftextcolors);
  FTextVertices:=copy(x.FTextVertices);

  FAreaBounderyCount:=x.FAreaBounderyCount;
  FAreaBounderies:=x.FAreaBounderies;
  FMarkedObjectCount :=x.FMarkedObjectCount;
  FMarkedObject:=x.FMarkedObject;
end;

procedure TOverlayLayerData.DrawEllipse(x, y, w, h: GLfloat; const aColor: TColor;
    const IsFilled: Boolean = False);
var
  i: Integer;
  degInRad1, degInRad2: GLfloat;
  xorg,yorg,wx,wy : glfloat;
begin
  xorg:=(x+(w-x)/2)/width;
  yorg:=( y+(h-y)/2)/height;
  wx:= 0.5*(w-x)/width;
  wy:= 0.5*(h-y)/height;

  degInRad1:=0;
  if IsFilled then
   begin
    for i := 1 to 360 do
      begin
        degInRad2:=degInRad1;
        degInRad1:=DegInRad1+Deg2Rad;
{        degInRad1 := i * DEG2RAD;
        degInRad2 := (i - 1) * DEG2RAD;}

        IntDrawTriangle(xorg + (Cos(degInRad1) * wx), yorg+ (Sin(degInRad1) * wy),
                       xorg + (Cos(degInRad2) * wx), yorg + (Sin(degInRad2) * wy),
                       xorg, yorg, aColor)
      end;
  end
  else
    for i := 1 to 360 do
      begin
        degInRad2:=degInRad1;
        degInRad1:=DegInRad1+Deg2Rad;
        intDrawLine(xorg + (Cos(degInRad1) * wx), yorg + (Sin(degInRad1) * wy),
                   xorg + (Cos(degInRad2) * wx), yorg + (Sin(degInRad2) * wy), aColor);
      end;
end;
                
procedure TOverlayLayerData.loadfromstream(t:tstream;vers:integer);
var n,i : integer;
begin
   t.read(frectcount,sizeof(glint));
   setlength(frects,frectcount);
   setlength(FRectColors,frectcount);
   if frectcount<>0 then
     begin
       t.read(frects[0],frectcount*sizeof(tglrect));
       t.read(frectcolors[0],frectcount*sizeof(tglrectcolor));
     end;

   t.read(ftrianglecount,sizeof(glint));
   setlength(FTriangles,frectcount);
   setlength(FTriangleColors,frectcount);
   if ftrianglecount<>0 then
     begin
       t.read(ftriangles[0],ftrianglecount*sizeof(tgltriangle));
       t.read(ftrianglecolors[0],ftrianglecount*sizeof(tgltrianglecolor));
     end;

   t.Read(flinecount,sizeof(glint));
   SetLength(FLines,FLineCount);
   SetLength(FLinecolors,FLineCount);
   if flinecount<>0 then
     begin
       t.read(flines[0],flinecount*sizeof(tglline));
       t.read(flinecolors[0],flinecount*sizeof(tgllinecolor));
     end;

   t.read(fThickLinecount,sizeof(glint));
   SetLength(FThickLines,FThickLineCount);
   SetLength(FThickLinecolors,FThickLineCount);
   if fThickLinecount<>0 then
     begin
       t.read(fThickLines[0],fthicklinecount*sizeof(tglLine));
       t.read(fThickLinecolors[0],fthicklinecount*sizeof(tglLinecolor));
     end;

   t.read(fStippleLinecount,sizeof(glint));
   SetLength(FStippleLines,FStippleLineCount);
   SetLength(FStippleLinecolors,FStippleLineCount);
   if fStippleLinecount<>0 then
     begin
       t.read(fStippleLines[0],fStipplelinecount*sizeof(tglLine));
       t.read(fStippleLinecolors[0],fStipplelinecount*sizeof(tglLinecolor));
     end;

   FTexts.Clear;
   t.read(n,sizeof(n));
   for i:=0 to n-1 do
     ftexts.Add(ldstr(t));

   setlength(ftextcolors,n);
   setlength(fTextvertices,n);
   if n<>0 then
     begin
       t.read(ftextcolors[0],n*sizeof(tglcolor));
       t.read(fTextvertices[0],n*sizeof(tglvertex));
     end;
     {                    FAreaBounderyCount : integer;
                    FAreaBounderies : array[0..maxRegions-1] of TOverlayObjectArray;
                    FMarkedObjectCount : integer;
                    FMarkedObject   : array[0..MaxMarkedObjects-1] of TMarkedObject;

}
  fmarkedobjectcount:=0;
  if vers>=$10002 then
    begin
      t.read(FAreaBounderyCount,sizeof(integer));
      if FAreaBounderyCount>0 then
        t.read(FAreaBounderies[0],sizeof(toverlayobjectarray)*fareabounderycount);
     { t.read(fmarkedobjectcount,sizeof(integer));
      if FMarkedObjectCount>0 then
        t.Read(fmarkedobject[0],sizeof(TMarkedObject)*FMarkedObjectCount);}
    end;
end;

procedure TOverlayLayerData.savetostream(t:tstream);
var n,i : integer;
begin
   t.Write(frectcount,sizeof(glint));
   if frectcount<>0 then
     begin
       t.Write(frects[0],frectcount*sizeof(tglrect));
       t.Write(frectcolors[0],frectcount*sizeof(tglrectcolor));
     end;

   t.Write(ftrianglecount,sizeof(glint));
   if ftrianglecount<>0 then
     begin
       t.Write(ftriangles[0],ftrianglecount*sizeof(tgltriangle));
       t.Write(ftrianglecolors[0],ftrianglecount*sizeof(tgltrianglecolor));
     end;

   t.Write(flinecount,sizeof(glint));
   if flinecount<>0 then
     begin
       t.Write(flines[0],flinecount*sizeof(tglline));
       t.Write(flinecolors[0],flinecount*sizeof(tgllinecolor));
     end;

   t.Write(fThickLinecount,sizeof(glint));
   if fThickLinecount<>0 then
     begin
       t.Write(fThickLines[0],fthicklinecount*sizeof(tglLine));
       t.Write(fThickLinecolors[0],fthicklinecount*sizeof(tglLinecolor));
     end;

   t.Write(fStippleLinecount,sizeof(glint));
   if fStippleLinecount<>0 then
     begin
       t.Write(fStippleLines[0],fStipplelinecount*sizeof(tglLine));
       t.Write(fStippleLinecolors[0],fStipplelinecount*sizeof(tglLinecolor));
     end;

   n:=FTexts.Count;
   t.write(n,sizeof(n));
   for i:=0 to n-1 do
     wrstr(t,ftexts[i]);

   n:=ftexts.count;
   if n<>0 then
     begin
       t.Write(ftextcolors[0],n*sizeof(tglcolor));
       t.Write(fTextvertices[0],n*sizeof(tglvertex));
     end;

   t.write(FAreaBounderyCount,sizeof(integer));
   if FAreaBounderyCount>0 then
     t.write(FAreaBounderies[0],sizeof(toverlayobjectarray)*fareabounderycount);
{   t.write(fmarkedobjectcount,sizeof(integer));
   if FMarkedObjectCount>0 then
     t.write(fmarkedobject[0],sizeof(TMarkedObject)*FMarkedObjectCount);}
end;

procedure TOverlayLayerData.DrawLine(x1, y1, x2, y2: GLfloat; const aColor: TColor);
var
  Index: Integer;
begin
  if FLineCount + 1 > Length(FLines) then
    InflateLines;

  Index := FLineCount;
  Inc(FLineCount); // common for rects and rectcolors!

  // vertices

  FLines[Index][1][1] := x1/width;
  FLines[Index][1][2] := y1/height;

  FLines[Index][2][1] := x2/width;
  FLines[Index][2][2] := y2/height;

  // colors

  pcolor(@FLinecolors[index][1][1])^:=acolor;
  pcolor(@FLinecolors[index][2][1])^:=acolor;
end;

procedure TOverlayLayerData.intDrawLine(x1, y1, x2, y2: GLfloat; const aColor: TColor);
var
  Index : Integer;
begin
  if FLineCount + 1 > Length(FLines) then
    InflateLines;

  Index := FLineCount;
  Inc(FLineCount); // common for rects and rectcolors!

  // vertices

  FLines[Index][1][1] := x1;
  FLines[Index][1][2] := y1;

  FLines[Index][2][1] := x2;
  FLines[Index][2][2] := y2;

  // colors

  pcolor(@FLinecolors[index][1][1])^:=acolor;
  pcolor(@FLinecolors[index][2][1])^:=acolor;
end;


procedure TOverlayLayerData.IntDrawTriangle(x1, y1, x2, y2, x3, y3: GLfloat;
  const aColor: TColor);
var
  Index: Integer;

begin
  if FTriangleCount + 1 > Length(FTriangles) then
    InflateTriangles;

  Index := FTriangleCount;
  Inc(FTriangleCount); // common for rects and rectcolors!

  // vertices

  FTriangles[Index][1][1] := x1;
  FTriangles[Index][1][2] := y1;

  FTriangles[Index][2][1] := x2;
  FTriangles[Index][2][2] := y2;

  FTriangles[Index][3][1] := x3;
  FTriangles[Index][3][2] := y3;

  // colors
  pcolor(@FTriangleColors[index][1][1])^:=acolor;
  pcolor(@FTriangleColors[index][2][1])^:=acolor;
  pcolor(@FTriangleColors[index][3][1])^:=acolor;
end;

procedure TOverlayLayerData.DrawRect(x1, y1, x2, y2: GLfloat; const aColor: TColor;
  const IsFilled: Boolean);
var
  Index: Integer;
begin

  if not IsFilled then
  begin
    DrawLine(x1, y1, x2, y1, aColor); // definitely could use inline
    DrawLine(x2, y1, x2, y2, aColor);
    DrawLine(x2, y2, x1, y2, aColor);
    DrawLine(x1, y2, x1, y1, aColor);
    Exit;
  end;

  if FRectCount + 1 > Length(FRects) then
    InflateRects;

  Index := FRectCount;
  Inc(FRectCount); // common for rects and rectcolors!

  // vertices

  FRects[Index][1][1] := x1 /width;
  FRects[Index][1][2] := y1 /height;

  FRects[Index][2][1] := x2 /width;
  FRects[Index][2][2] := y1 /height;

  FRects[Index][3][1] := x2 /width;
  FRects[Index][3][2] := y2 /height;

  FRects[Index][4][1] := x1/width;
  FRects[Index][4][2] := y2/height;

  // colors

  pcolor(@FRectColors[index][1][1])^:=acolor;
  pcolor(@FRectColors[index][2][1])^:=acolor;
  pcolor(@FRectColors[index][3][1])^:=acolor;
  pcolor(@FRectColors[index][4][1])^:=acolor;
end;

procedure TOverlayLayerData.DrawThickLine(x1, y1, x2, y2: GLfloat;
  const aColor: TColor);
var
  Index : Integer;
begin

  if FThickLineCount + 1 > Length(FThickLines) then
    InflateThickLines;

  Index := FThickLineCount;
  Inc(FThickLineCount); // common for rects and rectcolors!

  FThickLines[Index][1][1] := x1/width;
  FThickLines[Index][1][2] := y1/height;

  FThickLines[Index][2][1] := x2/width;
  FThickLines[Index][2][2] := y2/height;

  // colors

  pcolor(@FThickLinecolors[index][1][1])^:=acolor;
  pcolor(@FThickLinecolors[index][2][1])^:=acolor;
end;

procedure TOverlayLayerData.DrawStippleLine(x1, y1, x2, y2: GLfloat;
  const aColor: TColor);
var
  Index : Integer;
begin
  if FStippleLineCount + 1 > Length(FStippleLines) then
    InflateStippleLines;

  Index := FStippleLineCount;
  Inc(FStippleLineCount); // common for rects and rectcolors!

  FStippleLines[Index][1][1] := x1/width;
  FStippleLines[Index][1][2] := y1/height;

  FStippleLines[Index][2][1] := x2/width;
  FStippleLines[Index][2][2] := y2/height;

  // colors

  pcolor(@FStippleLinecolors[index][1][1])^:=acolor;
  pcolor(@FStippleLinecolors[index][2][1])^:=acolor;
end;


procedure TOverlayLayerData.DrawTriangle(x1, y1, x2, y2, x3, y3: GLfloat;
  const aColor: TColor; const IsFilled: Boolean = False);
var
  Index : Integer;
begin

  if not IsFilled then
  begin
    DrawLine(x1, y1, x2, y2, aColor); // definitely could use inline
    DrawLine(x2, y2, x3, y3, aColor);
    DrawLine(x3, y3, x1, y1, aColor);
    Exit;
  end;

{  if FListStatus <> glsFilling then
    ClearFlush; // set to filling and start timer, no other need here, except perhaps makecurrent
}
  if FTriangleCount + 1 > Length(FTriangles) then
    InflateTriangles;

  Index := FTriangleCount;
  Inc(FTriangleCount); // common for rects and rectcolors!

  // vertices

  FTriangles[Index][1][1] := x1/width;
  FTriangles[Index][1][2] := y1/height;

  FTriangles[Index][2][1] := x2/width;
  FTriangles[Index][2][2] := y2/height;

  FTriangles[Index][3][1] := x3/width;
  FTriangles[Index][3][2] := y3/height;

  // colors
  pcolor(@FTriangleColors[index][1][1])^:=acolor;
  pcolor(@FTriangleColors[index][2][1])^:=acolor;
  pcolor(@FTriangleColors[index][3][1])^:=acolor;
end;

procedure TOverlayLayerData.InflateLines;
const
  DEF_SIZE = 5000;
begin
  SetLength(FLines, Length(FLines) + DEF_SIZE);
  SetLength(FLineColors, Length(FLineColors) + DEF_SIZE);
end;

procedure TOverlayLayerData.InflateRects;
const
  DEF_SIZE = 10000;
begin
  SetLength(FRects, Length(FRects) + DEF_SIZE);
  SetLength(FRectColors, Length(FRectColors) + DEF_SIZE);
end;

procedure TOverlayLayerData.InflateThickLines;
const
  DEF_SIZE = 1000;
begin
  SetLength(FThickLines, Length(FThickLines) + DEF_SIZE);
  SetLength(FThickLineColors, Length(FThickLineColors) + DEF_SIZE);
end;

procedure TOverlayLayerData.InflateStippleLines;
const
  DEF_SIZE = 1000;
begin
  SetLength(FStippleLines, Length(FStippleLines) + DEF_SIZE);
  SetLength(FStippleLineColors, Length(FStippleLineColors) + DEF_SIZE);
end;

procedure TOverlayLayerData.InflateTriangles;
const
  DEF_SIZE = 5000; // 360 per ellipse.. so lets be bigger
begin
  SetLength(FTriangles, Length(FTriangles) + DEF_SIZE);
  SetLength(FTriangleColors, Length(FTriangleColors) + DEF_SIZE);
end;

constructor TOverlayLayerData.Create;
begin
  FTexts := TStringList.Create;
  {end inherited}

  FRectCount := 0;
  FLineCount := 0;
  FThickLineCount := 0;
  FStippleLineCount := 0;
  SetLength(FRects, 0);
  SetLength(FRectColors, 0);
  SetLength(FLines, 0);
  SetLength(FLineColors, 0);
  SetLength(FThickLines, 0);
  SetLength(FThickLineColors, 0);
  SetLength(FStippleLines, 0);
  SetLength(FStippleLineColors, 0);
  FAreaBounderyCount :=0;
  FMarkedObjectCount :=0;
end;

destructor TOverlayLayerData.Destroy;
begin
  freeandnil(ftexts);
  inherited;
end;

{ TOverlayData }

constructor TOverlayData.Create;
var i : integer;
begin
  for i:=low(layers) to high(layers) do
    layers[i]:=TOverlayLayerData.Create;
 fblobcount:=0;   
end;

procedure TOverlayData.Clear;
var i : integer;
begin
  for i:=low(layers) to high(layers) do
    layers[i].clear;
  fblobcount:=0;    
end;

procedure TOverlayData.assign(x:TOverlaydata);
var i : integer;
begin
  for i:=low(layers) to high(layers) do
    layers[i].assign(x.layers[i]);
  fblobcount:=x.fblobcount;
  fblobrects:=x.fblobrects;
  fbloblevel:=x.fbloblevel;
end;

{
    TGLOverlayHeader = record
                       magic   : integer;
                       hdrsize : integer;
                       version : integer;
                       layers  : integer;
                       end;
}

procedure TOverlayData.savetostream(t:tstream);
var  i : integer;
     hdr : TGLOverLayHeader;
begin
  hdr.version:=GLOverlayVersion;
  hdr.hdrsize:=sizeof(hdr);
  hdr.layers:=nrOverlayLayers;
  hdr.magic:=GLOverlaymagic;
  t.Write(hdr,sizeof(hdr));
  t.write(fwidth,sizeof(fwidth));
  t.write(fheight,sizeof(fheight));

  t.write(FBlobCount,sizeof(integer));
  if FBlobCount>0 then
     begin
       t.write(FBlobRects[0],sizeof(TRect)*fblobcount);
       t.write(FBloblevel[0],sizeof(integer)*fblobcount);
     end;
  for i:=low(layers) to high(layers) do
    layers[i].savetostream(t);
end;

procedure TOverlayData.loadfromstream(t:tstream);
var i,k : integer;
    hdr : TGLOverLayHeader;
    emptybuf : array[0..49] of char;

begin
  t.Read(hdr,sizeof(integer)*4);
  i:=hdr.hdrsize-4*sizeof(integer); // hdrsize in file minus read parts.
  k:=sizeof(hdr)-4*sizeof(integer); // hdrsize according to structures in src, minus. ..
  if i>k then
    begin
      if k<>0 Then
        t.read(hdr.theend,k);
      while i>0 do
        begin
          k:=i;
          if i>sizeof(emptybuf) then
            k:=sizeof(emptybuf);
          t.read(emptybuf,k);
          dec(i,k);
        end;
    end
  else
    if i<>0 then
       t.read(hdr.theend,i);
  t.read(fwidth,sizeof(fwidth));
  t.read(fheight,sizeof(fheight));

  if hdr.version>=$10002 then
    begin
      t.read(FBlobCount,sizeof(integer));
      if FBlobCount>0 then
        begin
          t.read(FBlobRects[0],sizeof(TRect)*fblobcount);
          t.read(FBloblevel[0],sizeof(integer)*fblobcount);
        end;
    end;
  for i:=1 to hdr.layers do
    begin
      layers[i].width:=fwidth;
      layers[i].height:=fheight;
      layers[i].loadfromstream(t,hdr.version);
    end;
end;


destructor TOverlayData.Destroy;
var i : integer;
begin
  for i:=low(layers) to high(layers) do
    freeandnil(layers[i]);
  inherited;
end;

procedure TOverlayData.setwidth(w:integer);
var i : integer;
begin
 fwidth:=w;
  for i:=low(layers) to high(layers) do
    layers[i].width:=w;
end;

procedure TOverlayData.setheight(h:integer);
var i : integer;
begin
 fheight:=h;
  for i:=low(layers) to high(layers) do
    layers[i].height:=h;
end;
procedure TOverlayData.markblob(x1, y1, x2, y2, level: integer);
begin
 if fblobcount<maxblobrect then
  begin
    with fblobrects[fblobcount] do
     begin
      Left:=x1; top:=y1;
      right:=x2; bottom:=y2;
     end;
    fbloblevel[fblobcount]:=level;
    inc(fblobcount); 
  end;
end;


procedure TOverlayLayerData.markobject(t: TOverlayObjectType; nr:integer=1);
var ind : integer;
begin
  if FMarkedObjectCount<maxmarkedobjects then
    begin
      ind:=-1;
      case t of
        ovlRect      : begin
                         ind:=FLineCount;
                         nr:=4*nr;
                       end;
        ovlRectfilled: ind:=FRectCount;
        ovlLine      : ind:=FLineCount;
        ovlThickline : ind:=FThickLineCount;
        ovlTriangle  : ind:=FTriangleCount;
        ovlStippleline : ind:=FStippleLineCount;
        ovlCirclefilled: ind:=FTriangleCount;
        ovlCircleEmpty : begin
                           ind:=FLineCount;
                           nr:=nr*360;
                         end;
        end;
      if ind<>-1 then
        begin
          FMarkedObject[FMarkedObjectCount].objindex:=ind;
          FMarkedObject[FMarkedObjectCount].objtype:=t;
          FMarkedObject[FMarkedObjectCount].number:=nr;
          inc(FMarkedObjectCount);
        end;
    end;
end;

procedure xorcolor(p:pcardinal;nr:integer);
var p2:PCardinal;
begin
  p2:=@pculongarray(p)[nr];
  while (pbbyte(p)<pbbyte(p2)) do
    begin
      p^:=p^ or cardinal(-1); //$50000000;
      inc(p);
    end;
end;

procedure TOverlaylayerData.colorselection(col:tcolor);
{$ifdef newselected}
var
    i,j : integer;
    t : TOverlayObjectType;
    nr:integer;
    ind:integer;
{$endif}
begin
  {$ifdef newselected}
  if fmarkedobjectcount>0 then
   begin
     for i:=0 to FMarkedObjectCount-1 do
       begin
         ind:=FMarkedObject[i].objindex;
         nr:=FMarkedObject[i].number;
         t:=FMarkedObject[i].objtype;
         case t of
           ovlline:  begin
                        if (ind+nr)<=FLineCount then
                           xorcolor(@flinecolors[ind],nr*2);               // 2 vertices per line
                     end;
           ovlRect:  begin
                       if (ind+nr)<=FLineCount then
                           xorcolor(@flinecolors[ind],nr*2);               // 2 vertices per line
                     end;
           ovlCircleEmpty:  begin
                       if (ind+nr)<=FLineCount then
                           xorcolor(@flinecolors[ind],nr*2);               // 360 lines per circle
                     end;
             end;

       end;
   end;
  {$endif}
end;

function TOverlayData.getblob(ind:integer):trect;
begin
  result:=fblobrects[ind];
end;

end.

